/*  Emacs style mode select   -*- C++ -*-  */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id:$ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  DESCRIPTION: */
/* 	Fixed point arithemtics, implementation. */
/*  */
/* ----------------------------------------------------------------------------- */


#ifndef __M_FIXED__
#define __M_FIXED__


#ifdef __GNUG__
#pragma interface
#endif


/*  */
/*  Fixed point, 32bit as 16.16. */
/*  */
#define FRACBITS		16
#define FRACUNIT		(1<<FRACBITS)

typedef int fixed_t;	/* NB: It is a BUG to assume 'int' same as 'long' */
						/* but it is too big a hassle to edit everywhere. */
						/* So for the present I'll accept this bad usage. */

#if		(defined(__GNUC__) && defined(__M68020__) )

/*RA: The two 'extended asm' macros below allow  fixnumber calculations		*/
/*    to be performed without function call overhead costs.	The only cost	*/
/*    involved, apart from pure calculation, is possible typecasting, but	*/
/*    that cost is inevitable when needed, as for normal function calls.	*/
/*    But in most cases even typecasting is cost free, as the compiler can	*/
/*    optimize it away.  But the compiler optimizer needs 'free hands', so	*/
/*    we must help keep choices open, as shown in some comments below.		*/

#define	FixedMul(x,y)														\
__extension__					/* skip 'pedantic' compiler warnings	*/	\
({	register fixed_t retv;		/* leave register choice to optimizer	*/	\
	__asm__						/* avoid 'volatile' to help optimizer	*/	\
	("	mulsl	%2,d1:%0;		/* calc d1 = p3.p2,  %0 = p1.p0			*/	\
		movew	d1,%0;			/* calc %0 = p1.p2						*/	\
		swap	%0				/* calc %0 = p2.p1,  the fixed product	*/	\
	"				/* end of code */										\
	:	"=d"	(retv)			/* 0 'retv' output needs a 'd' register	*/	\
	:	"0"		((fixed_t)(x))	/* 1 make input 'x' use 'retv' reg		*/	\
	,	"imd"	((fixed_t)(y))	/* 2 input _y is any data (no areg)		*/	\
	:	"cc","d1"				/* clobber cc,d1						*/	\
	);				/* end of asm */										\
	retv;						/* make 'retv' current compiler value	*/	\
})

/* Simple overflows can occur for large values. At present it is ignored.	*/
/* But saved execution time would be enough to handle it at present speed.	*/

/* In the FixedDiv routine below, however, overflow and undefined results	*/
/* can occur (as always with division).  That may cause 'div by zero' traps	*/
/* and the implementation should not depend on the trap handler of the OS	*/
/* to deal with that gracefully.  OS implementations of such features vary.	*/
/* For the present, the FixedDiv code ignores all such problems...			*/

#define	FixedDiv(x,y)														\
__extension__					/* skip 'pedantic' compiler warnings	*/	\
({	register fixed_t retv;		/* leave register choice to optimizer	*/	\
	__asm__																	\
	("	swap	%0;				/* calc %0 = divdlo.divdhi				*/	\
		movew	%0,d1;			/* calc d1 = ?.divdhi					*/	\
		extl	d1;				/* calc	d1 = signext.divdhi				*/	\
		clrw	%0;				/* calc %0 = divdlo.zero				*/	\
		divsl	%2,d1:%0		/* calc d1=remainder  %0 = quotient		*/	\
	"				/* end of code */										\
	:	"=d"	(retv)			/* 0 'retv' output needs a 'd' register	*/	\
	:	"0"		((fixed_t)(x))	/* 1 make input 'x' use 'retv' reg		*/	\
	,	"imd"	((fixed_t)(y))	/* 2 input _y is any data (no areg)		*/	\
	:	"cc","d1"				/* clobber cc,d1						*/	\
	);				/* end of asm */										\
	retv;						/* make 'retv' current compiler value	*/	\
})

/* The macros above should give a significant speed increase for fixnumber	*/
/* calculations, but another option is to use that gain to add safety check	*/
/* of input data.  That is one way to eliminate the 'div by zero' problem.	*/
/* But for the present I choose to keep the speed gains as such...			*/

#else	/* not Atari GnuC */

fixed_t FixedMul	(fixed_t a, fixed_t b);
fixed_t FixedDiv	(fixed_t a, fixed_t b);
fixed_t FixedDiv2	(fixed_t a, fixed_t b);

#endif	/* ATARI __GNUC__ */

#endif	/* __M_FIXED__ */
