/*  Emacs style mode select   -*- C++ -*-  */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id:$ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  $Log:$ */
/*  */
/*  DESCRIPTION: */
/*  */
/* ----------------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <stdarg.h>
#include <sys/time.h>
#include <unistd.h>

#include "doomdef.h"
#include "m_misc.h"
#include "i_video.h"
#include "i_sound.h"
#include "i_net.h"

#include "d_net.h"
#include "g_game.h"
#include "m_argv.h"

#include "snd_server.h"
#include "vid_svgalib.h"
#include "vid_x11.h"
#include "snd_linux.h"
#include "net_unix.h"

#ifdef __GNUG__
#pragma implementation "i_system.h"
#endif
#include "i_system.h"

int	mb_used = 8;	/*  8 Mo par dfaut */

void I_NullSysFunc(void);

void
I_Tactile
( int	on,
  int	off,
  int	total )
{
  /*  UNUSED. */
  on = off = total = 0;
}

ticcmd_t	emptycmd;
ticcmd_t*	I_BaseTiccmd(void)
{
	return &emptycmd;
}

int  I_GetHeapSize (void)
{
	return mb_used*1024*1024;
}

byte* I_ZoneBase (int*	size)
{
	*size = mb_used*1024*1024;

	printf(" %d Kbytes allocated for zone\n",(*size)>>10);

	return (byte *) malloc (*size);
}



/*  */
/*  I_GetTime */
/*  returns time in 1/70th second tics */
/*  */

int  I_GetTime (void)
{
    static int		basetime=0;
    int			newtics;
    struct timeval	tp;
    struct timezone	tzp;
  
    gettimeofday(&tp, &tzp);
    if (!basetime)
	basetime = tp.tv_sec;
    newtics = (tp.tv_sec-basetime)*TICRATE + tp.tv_usec*TICRATE/1000000;

    return newtics;
}


/*  */
/*  I_Init */
/*  */
void I_Init (void)
{
	I_InitSound();
	/* I_InitGraphics(); */
	I_InitControl();

	signal(SIGINT, (void (*)(int)) I_Quit);
}

/*  */
/*  I_Quit */
/*  */
void I_Quit (void)
{
	D_QuitNetGame ();
	I_ShutdownNetwork();
	I_ShutdownControl();
	I_ShutdownSound();
	I_ShutdownMusic();
	M_SaveDefaults ();
	I_ShutdownGraphics();
	exit(0);
}

void I_WaitVBL(int count)
{
#ifdef SGI
    sginap(1);                                           
#else
#ifdef SUN
    sleep(0);
#else
    usleep (count * (1000000/70) );                                
#endif
#endif
}

byte*	I_AllocLow(int length)
{
    byte*	mem;
        
    mem = (byte *)malloc (length);
    memset (mem,0,length);
    return mem;
}


/*  */
/*  I_Error */
/*  */
extern boolean demorecording;

void I_Error (char *error, ...)
{
    va_list	argptr;

    /*  Message first. */
    va_start (argptr,error);
    fprintf (stderr, "Error: ");
    vfprintf (stderr,error,argptr);
    fprintf (stderr, "\n");
    va_end (argptr);

    fflush( stderr );

	/*  Shutdown. Here might be other errors. */
	if (demorecording)
		G_CheckDemoStatus();

	D_QuitNetGame ();
	I_ShutdownNetwork();
	I_ShutdownControl();
	I_ShutdownSound();
	I_ShutdownMusic();
	I_ShutdownGraphics();
    
	exit(-1);
}

void I_StartTic(void)
{
	I_UpdateKeyboard();
	I_UpdateMouse();
	I_UpdateJoystick();
}

/* ----------------------------------------------------------------- */
/*                 Dynamic system dependent functions                */
/* ----------------------------------------------------------------- */

typedef struct {
	char *name;	/* Routine name */
	char *param;	/* Parameter to force the use of this routine */
	int number;	/* Number of the routine */
} routlist_t;

enum {
	SYSROUT_OFF =0,
	SYSROUT_ON,
	SYSROUT_SVGALIB,
	SYSROUT_X11,
	SYSROUT_SNDSERV,
	SYSROUT_SNDLINUX
};

routlist_t system_rout[]={
	{"Off",			"off",		SYSROUT_OFF},
	{"On",			"on",		SYSROUT_ON},
	{"Svgalib",		"svgalib",	SYSROUT_SVGALIB},
	{"X11",			"x11",		SYSROUT_X11},
	{"Sound server",	"sndserver",	SYSROUT_SNDSERV},
	{"Linux /dev/audio",	"devaudio",	SYSROUT_SNDLINUX}
};


void I_SysInit(void)
{
	int p;

	int numkbd,nummouse,numjoystick;
	int numsound,numvideo;

	/* --- Initialise default system functions --- */

	/* controllers */

	I_InitControl = I_NullSysFunc; /* Init controller functions */
	I_ShutdownControl = I_NullSysFunc; /* Shutdown controller functions */

	I_UpdateKeyboard = I_NullSysFunc;
	numkbd = SYSROUT_OFF;
	I_UpdateMouse = I_NullSysFunc;
	nummouse = SYSROUT_OFF;
	I_UpdateJoystick = I_NullSysFunc;
	numjoystick = SYSROUT_OFF;

	/* sound */

	I_SndInit = I_NullSysFunc;
	I_SndUpdate = I_NullSysFunc;
	I_InitSound = I_NullSysFunc;
	I_ShutdownSound = I_NullSysFunc;
	I_StartSound = I_NullSysFunc;
	numsound = SYSROUT_OFF;

	/* video */

	I_VidInit = I_NullSysFunc;
	I_VidUpdate = I_NullSysFunc;
	I_SetPalette256 = I_NullSysFunc;
	I_InitGraphics = I_InitGraphics_null;
	I_ShutdownGraphics = I_NullSysFunc;
	I_SetMainScreen = I_NullSysFunc;
	numvideo = SYSROUT_OFF;

	/* network */

	I_InitNetwork = I_InitNetwork_unix;
	I_ShutdownNetwork = I_ShutdownNetwork_unix;

	/* --- Initialise default LINUX functions --- */

	if (getenv("DISPLAY"))
	{
		/* X11 controllers */
		I_UpdateKeyboard = I_UpdateKeyboard_x11;
		numkbd = SYSROUT_X11;
		I_UpdateMouse = I_UpdateMouse_x11;
		nummouse = SYSROUT_X11;

		/* X11 display */
		I_VidUpdate = I_VidUpdate_x11;
		I_SetPalette256 = I_SetPalette256_x11;
		I_InitGraphics = I_InitGraphics_x11;
		I_ShutdownGraphics = I_ShutdownGraphics_x11;
		numvideo=SYSROUT_X11;
	} else {
		/* Svgalib controllers */
		I_UpdateKeyboard = I_UpdateKeyboard_svgalib;
		numkbd = SYSROUT_SVGALIB;

		/* Svgalib display */
		I_VidInit = I_VidInit_svgalib;
		I_VidUpdate = I_VidUpdate_svgalib;
		I_SetPalette256 = I_SetPalette256_svgalib;
		I_InitGraphics = I_InitGraphics_svgalib;
		I_ShutdownGraphics = I_ShutdownGraphics_svgalib;
		numvideo = SYSROUT_SVGALIB;
	}

	I_SndInit = I_NullSysFunc;
	I_SndUpdate = I_NullSysFunc;
	I_InitSound = I_InitSound_sndserver;
	I_ShutdownSound = I_ShutdownSound_sndserver;
	I_StartSound = I_StartSound_sndserver;
	numsound=SYSROUT_SNDSERV;

	/* --- Initialise forced functions --- */

	p=M_CheckParm("-keyboard");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			/* Keep keyboard routine under X11 */
			if ( I_UpdateKeyboard != I_UpdateKeyboard_x11)
			{
				I_UpdateKeyboard = I_NullSysFunc;
				numkbd=SYSROUT_OFF;
			}
		}		
	}

	p=M_CheckParm("-mouse");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateMouse = I_NullSysFunc;
			nummouse=SYSROUT_OFF;
		}		
	}

	p=M_CheckParm("-joystick");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateJoystick = I_NullSysFunc;
			numjoystick=SYSROUT_OFF;
		}		
	}

	p=M_CheckParm("-sound");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_SndInit = I_NullSysFunc;
			I_SndUpdate = I_NullSysFunc;
			I_InitSound = I_NullSysFunc;
			I_ShutdownSound = I_NullSysFunc;
			numsound=SYSROUT_OFF;
		}		
		if (!strcmp(system_rout[SYSROUT_SNDLINUX].param,myargv[p+1])) {
			I_SndInit = I_NullSysFunc;
			I_SndUpdate = I_SndUpdate_linux;
			I_InitSound = I_InitSound_linux;
			I_ShutdownSound = I_ShutdownSound_linux;
			I_StartSound = I_StartSound_linux;
			numsound = SYSROUT_SNDLINUX;
		}
		if (!strcmp(system_rout[SYSROUT_SNDSERV].param,myargv[p+1])) {
			I_SndInit = I_NullSysFunc;
			I_SndUpdate = I_NullSysFunc;
			I_InitSound = I_InitSound_sndserver;
			I_ShutdownSound = I_ShutdownSound_sndserver;
			I_StartSound = I_StartSound_sndserver;
			numsound=SYSROUT_SNDSERV;
		}		
	}

	p=M_CheckParm("-video");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			/* Disable keyboard,mouse if no appropriate video */
			if (( I_VidUpdate == I_VidUpdate_x11) ||
				(I_VidUpdate == I_VidUpdate_svgalib))
			{
				I_UpdateKeyboard = I_NullSysFunc;
				numkbd=SYSROUT_OFF;
				I_UpdateMouse = I_NullSysFunc;
				nummouse=SYSROUT_OFF;
			}

			I_VidInit = I_NullSysFunc;
			I_VidUpdate = I_NullSysFunc;
			I_SetPalette256 = I_NullSysFunc;
			I_InitGraphics = I_InitGraphics_null;
			I_ShutdownGraphics = I_NullSysFunc;
			I_SetMainScreen = I_NullSysFunc;
			numvideo=SYSROUT_OFF;
		}		
	}

	/* Display routines names */

	printf("Keyboard.: %s\n",system_rout[numkbd].name);
	printf("Mouse....: %s\n",system_rout[nummouse].name);
	printf("Joystick.: %s\n",system_rout[numjoystick].name);
	printf("Sound....: %s\n",system_rout[numsound].name);
	printf("Video....: %s\n\n",system_rout[numvideo].name);

	/* First initializations */

	I_VidInit();
	I_SndInit();
}

/* --- Default system dynamic function = Null */

void I_NullSysFunc(void)
{
}

/* --- Controllers : update (keyboard,mouse,joystick,joypads,etc... */

void (*I_InitControl) (void);
void (*I_ShutdownControl) (void);

void (*I_UpdateKeyboard) (void);
void (*I_UpdateMouse) (void);
void (*I_UpdateJoystick) (void);

/* --- Sound */

void (*I_SndInit)(void);		/* Initializations required for sound */
void (*I_SndUpdate)(void);		/* Update sound */
void (*I_InitSound)(void);		/* Initialize sound */
void (*I_ShutdownSound)(void);		/* Shutdown sound */

/* --- Music */

/* --- Video */

void (*I_VidInit)(void);		/* Initializations required for video */
void (*I_VidUpdate)(void);		/* Render the final screen */
void (*I_SetPalette256)(byte *palette); /* set a 256 colours palette */
void (*I_InitGraphics)(void);		/* Init video */
void (*I_ShutdownGraphics)(void);	/* Shutdown video */
void (*I_SetMainScreen)(void);		/* Set final screen pointer */
