/*
 *	GEM functions
 *
 *  Olivier Landemarre
 *	Johan Klockars
 *	Xavier Joubert
 *	Patrice Mandin
 */

#include <stdlib.h>

#include <mint/osbind.h>
#include <gem.h>

#include "d_main.h"
#include "d_event.h"
#include "doomdef.h"
#include "doomstat.h"
#include "v_video.h"
#include "m_argv.h"

#include "r_draw.h"
#include "am_map.h"
#include "f_wipe.h"
#include "f_finale.h"

#include "i_version.h"
#include "i_zoom.h"
#include "i_system.h"
#include "i_video.h"
#include "vid_gem.h"
#include "in_scancode.h"

#define DESK_HANDLE 0
#define S_ONLY 3

enum {
	MX_STRAM=0,
	MX_TTRAM,
	MX_PREFSTRAM,
	MX_PREFTTRAM
};

extern byte     *orig_colormaps;

static int      ap_id=-1;
static int      vdi_handle=0;
static void     *buffer;

static int		old_pal[256*3];
static unsigned char    vdi_index[256] = {0, 2, 3, 6, 4, 7, 5, 8, 9, 10, 11, 14, 12, 15, 13, 255, 0};

static int		wnd_handle=-1;
static char		*wnd_name=NULL;
static unsigned char gem_curclavier[128];
static unsigned char gem_oldclavier[128];

int VideoInited_gem=0;

static int create_window(int width,int height);
static int do_messages(int *message, int *kc, int*simul_kbd);
static void do_keyboard(int kc, int ks);
static void wind_redraw(int winhandle, int *inside);
static void copie_windowvdi(int *rect);

void I_VidInit_gem(void)
{
	int bufwidth, bufheight;
	int i,vdipixelsize;
	int work_in[12], work_out[57], dummy;
	int AllocFastRam; /* Alloc preferrably from Fast Ram */
	int rgb[3];
	int screensize;

/*	_global[0] = 0;*/
	ap_id = appl_init();

	work_in[0]=Getrez()+2;
	for(i = 1; i < 10; i++)
		work_in[i] = 1;
	work_in[10] = 2;

	vdi_handle = graf_handle(&dummy, &dummy, &dummy, &dummy);
	v_opnvwk(work_in, &vdi_handle, work_out);

	if(vdi_handle==0)
		I_Error("Error initializing VDI\n");

	videowidth = work_out[0] + 1;
	videoheight = work_out[1] + 1;
	vq_extnd(vdi_handle, 1, work_out);
	bpp = work_out[4];

	/* --- Allocate screens */
/*
	if ((videowidth<SCREENWIDTH) || (videoheight<SCREENHEIGHT))
		zoomscreen=true;
*/
	switch(bpp)
	{
		case 8 :
			vdipixelsize=1;
			break;
		case 15 :
		case 16 :
			vdipixelsize=2;
			break;
		case 24 :
			vdipixelsize=3;
			break;
		case 32 :
			vdipixelsize=4;
			break;
		default :
			vdipixelsize=0;
			break;
	}

	if (vdipixelsize==0)
		I_Error("You need a 8, 15, 16, 24 or 32 bits display!\n");

	/* No Zoom screen */

	bufwidth=SCREENWIDTH;
	bufheight=SCREENHEIGHT;

	/* Allocate buffer */

	screensize=bufwidth * bufheight * vdipixelsize;

	AllocFastRam=MX_PREFTTRAM;		/* Alloc from Fast Ram */
	if (M_CheckParm("-nofastblit"))
		AllocFastRam=MX_STRAM;	/* Alloc from normal Ram */

	if (mxalloc_present)
		buffer = (void *)Mxalloc(screensize, AllocFastRam);
	else
		buffer = (void *)Malloc(screensize);

	if (!buffer)
		I_Error("Not enough memory for draw buffer!\n");

	memset(buffer, 0,screensize);

	/* XJ: Save old palette */
	for(i = 0; i < 256; i++)
	{
		vq_color(vdi_handle, i, 0, rgb);
		old_pal[i*3+0] = rgb[0];
		old_pal[i*3+1] = rgb[1];
		old_pal[i*3+2] = rgb[2];
	}

	/* Create GEM window */
	wnd_handle=create_window(bufwidth,bufheight);	
	if (wnd_handle<0)
		I_Error("Can not create GEM window!\n");

	graf_mouse(ARROW, NULL);

	VideoInited_gem=1;
}

static int create_window(int width,int height)
{
#define WINDOW_TYPE (NAME|MOVER|CLOSER|SMALLER)
	int		x2,y2,w2,h2;
	int		desk_x,desk_y,desk_w,desk_h;
	int		posx,posy;

	/* Get desktop size and position */
	if (!wind_get(DESK_HANDLE,WF_WORKXYWH,&desk_x,&desk_y,&desk_w,&desk_h))
		return -1;

	/* Center our window */
	posx = desk_x;
	posy = desk_y;
	if (width<desk_w)
		posx += (desk_w - width) >> 1;
	if (height<desk_h)
		posy += (desk_h - height) >> 1;

	/* Calculate our window size and position */
	if (!wind_calc(0,WINDOW_TYPE,posx,posy,width,height,&x2,&y2,&w2,&h2))
		return -1;

	/* Create window */
	wnd_handle=wind_create(WINDOW_TYPE,x2,y2,w2,h2);
	if (wnd_handle<0)
		return wnd_handle;

	/* Set window name */
	wind_set(wnd_handle,WF_NAME,(int) version_name,0,0,0);

	/* Open the window */
	wind_open(wnd_handle,x2,y2,w2,h2);

	return wnd_handle;
}

void I_InitControl_gem(void)
{
	memset(gem_curclavier,0,sizeof(gem_curclavier));
	memset(gem_oldclavier,0,sizeof(gem_oldclavier));
}

void I_ShutdownGraphics_gem(void)
{	
	int i,rgb[3];
	
	/* Destroy GEM window */
	if (wnd_handle>=0) {
		wind_close(wnd_handle);
		wind_delete(wnd_handle);
		wnd_handle=-1;

		if (wnd_name) {
			free(wnd_name);
			wnd_name=NULL;
		}
	}

	/* XJ: Old palette */
	if (VideoInited_gem) {
		for(i = 0; i < 256; i++) {
			rgb[0] = old_pal[i*3+0];
			rgb[1] = old_pal[i*3+1];
			rgb[2] = old_pal[i*3+2];
			vs_color(vdi_handle, i, rgb);
		}
	}

	if (vdi_handle)
		v_clsvwk(vdi_handle);

	appl_exit();

	Mfree(buffer);
}

void I_VidUpdate_gem(void)
{
	int pxy[8];

	wind_get(wnd_handle, WF_WORKXYWH, &pxy[0], &pxy[1], &pxy[2], &pxy[3]);
	wind_redraw(wnd_handle, pxy);
}

void I_UpdateKeyboard_gem(void)
{
	for (;;)
	{
		int	buffer[8], neant, kc, ks, quit, resultat;

		ks = 0;
		
		resultat = evnt_multi(MU_MESAG|MU_TIMER|MU_KEYBD, 0,0,0, 0,0,0,0,0 ,0,0,0,0,0, buffer, 10L, &neant,&neant,&neant,&ks,&kc,&neant);

		if (resultat & MU_MESAG) {
			int simul_kbd;

			simul_kbd=0;

			quit = do_messages(buffer, &kc, &simul_kbd);
			if (quit) {
				I_Quit();
				break;
			}

			if (simul_kbd)
				resultat |= MU_KEYBD;
		}

		if (resultat & MU_KEYBD)
			do_keyboard(kc,ks);
		else
			do_keyboard(0,0);

		if (resultat & MU_TIMER)
			break;
	}
}

static int do_messages(int *message, int *kc, int *simul_kbd)
{
	int quit;

	quit = 0;
	switch (message[0]) {
		case WM_CLOSED:
			/* Simuler appui sur F10 pour quitter */
			*kc = SCANCODE_F10 << 8;
			*simul_kbd = 1;
			break;
		case WM_MOVED:
			wind_set(message[3],WF_CURRXYWH,message[4],message[5],message[6],message[7]);
			break;
		case WM_TOPPED:
			wind_set(message[3],WF_TOP,message[4],0,0,0);
			break;
		case WM_REDRAW:
			wind_redraw(message[3],&message[4]);
			break;
		case WM_ICONIFY:
		case WM_ALLICONIFY:
			wind_set(message[3],WF_ICONIFY,message[4],message[5],message[6],message[7]);
			/* Simuler appui sur pause */
			if (!paused) {
				*kc = SCANCODE_BREAK << 8;
				*simul_kbd = 1;
			}
			break;
		case WM_UNICONIFY:
			wind_set(message[3],WF_UNICONIFY,message[4],message[5],message[6],message[7]);
			break;
		case WM_BOTTOMED:
		case WM_UNTOPPED:
			/* Simuler appui sur pause */
			if (!paused) {
				*kc = SCANCODE_BREAK << 8;
				*simul_kbd = 1;
			}
			break;
		case AP_TERM:    
			quit = 1;
			break;
		default:
			break;
	}

	return quit;
}

static void do_keyboard(int kc, int ks)
{
	event_t event;	
	int		scancode,dummy;

	memset(gem_curclavier,0,128);
	
	if (kc) {
		scancode=(kc>>8) & 127;

		gem_curclavier[scancode]=0xFF;
	}

	if (!ks)
		graf_mkstate(&dummy, &dummy, &dummy, &ks);

	/* Read special keys */

	if (ks & K_RSHIFT)
		gem_curclavier[SCANCODE_RIGHTSHIFT]=0xFF;
	if (ks & K_LSHIFT)
		gem_curclavier[SCANCODE_LEFTSHIFT]=0xFF;
	if (ks & K_CTRL)
		gem_curclavier[SCANCODE_LEFTCONTROL]=0xFF;
	if (ks & K_ALT)
		gem_curclavier[SCANCODE_LEFTALT]=0xFF;

#define KEY_DOWN(a) (gem_curclavier[a])
#define KEY_PRESS(a) (gem_curclavier[a] && !gem_oldclavier[a])
#define KEY_WASPRESSED(a) (gem_oldclavier[a] && !gem_curclavier[a])
#define PROCESS_KEY(k,t)					\
	if (KEY_PRESS(k))					\
	{							\
			event.type = ev_keydown;		\
			event.data1 = t;			\
			D_PostEvent(&event);			\
	}							\
	if (KEY_WASPRESSED(k))					\
	{							\
			event.type = ev_keyup;			\
			event.data1 = t;			\
			D_PostEvent(&event);			\
	}

	PROCESS_KEY(SCANCODE_CURSORBLOCKRIGHT, KEY_RIGHTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKLEFT, KEY_LEFTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKDOWN, KEY_DOWNARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKUP, KEY_UPARROW);

	PROCESS_KEY(SCANCODE_ESCAPE, KEY_ESCAPE);
	PROCESS_KEY(SCANCODE_ENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_KEYPADENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_TAB, KEY_TAB);
	PROCESS_KEY(SCANCODE_SPACE, ' ');

	PROCESS_KEY(SCANCODE_F1, KEY_F1);
	PROCESS_KEY(SCANCODE_F2, KEY_F2);
	PROCESS_KEY(SCANCODE_F3, KEY_F3);
	PROCESS_KEY(SCANCODE_F4, KEY_F4);
	PROCESS_KEY(SCANCODE_F5, KEY_F5);
	PROCESS_KEY(SCANCODE_F6, KEY_F6);
	PROCESS_KEY(SCANCODE_F7, KEY_F7);
	PROCESS_KEY(SCANCODE_F8, KEY_F8);
	PROCESS_KEY(SCANCODE_F9, KEY_F9);
	PROCESS_KEY(SCANCODE_F10, KEY_F10);
	PROCESS_KEY(SCANCODE_F11, KEY_F11);
	PROCESS_KEY(SCANCODE_F12, KEY_F12);
		
	PROCESS_KEY(SCANCODE_1, '1');
	PROCESS_KEY(SCANCODE_2, '2');
	PROCESS_KEY(SCANCODE_3, '3');
	PROCESS_KEY(SCANCODE_4, '4');
	PROCESS_KEY(SCANCODE_5, '5');
	PROCESS_KEY(SCANCODE_6, '6');
	PROCESS_KEY(SCANCODE_7, '7');
	PROCESS_KEY(SCANCODE_8, '8');
	PROCESS_KEY(SCANCODE_9, '9');
	PROCESS_KEY(SCANCODE_0, '0');

	PROCESS_KEY(SCANCODE_MINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_KEYPADMINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_EQUAL, KEY_EQUALS);
	PROCESS_KEY(SCANCODE_KEYPADPLUS, KEY_EQUALS);

	PROCESS_KEY(SCANCODE_BACKSPACE, KEY_BACKSPACE);
	PROCESS_KEY(SCANCODE_BREAK, KEY_PAUSE);

	PROCESS_KEY(SCANCODE_LEFTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_RIGHTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_LEFTCONTROL, KEY_RCTRL);
	PROCESS_KEY(SCANCODE_LEFTALT, KEY_RALT);

	PROCESS_KEY(SCANCODE_Q, 'q');
	PROCESS_KEY(SCANCODE_W, 'w');
	PROCESS_KEY(SCANCODE_E, 'e');
	PROCESS_KEY(SCANCODE_R, 'r');
	PROCESS_KEY(SCANCODE_T, 't');
	PROCESS_KEY(SCANCODE_Y, 'y');
	PROCESS_KEY(SCANCODE_U, 'u');
	PROCESS_KEY(SCANCODE_I, 'i');
	PROCESS_KEY(SCANCODE_O, 'o');
	PROCESS_KEY(SCANCODE_P, 'p');

	PROCESS_KEY(SCANCODE_A, 'a');
	PROCESS_KEY(SCANCODE_S, 's');
	PROCESS_KEY(SCANCODE_D, 'd');
	PROCESS_KEY(SCANCODE_F, 'f');
	PROCESS_KEY(SCANCODE_G, 'g');
	PROCESS_KEY(SCANCODE_H, 'h');
	PROCESS_KEY(SCANCODE_J, 'j');
	PROCESS_KEY(SCANCODE_K, 'k');
	PROCESS_KEY(SCANCODE_L, 'l');

	PROCESS_KEY(SCANCODE_Z, 'z');
	PROCESS_KEY(SCANCODE_X, 'x');
	PROCESS_KEY(SCANCODE_C, 'c');
	PROCESS_KEY(SCANCODE_V, 'v');
	PROCESS_KEY(SCANCODE_B, 'b');
	PROCESS_KEY(SCANCODE_N, 'n');
	PROCESS_KEY(SCANCODE_M, 'm');

	memcpy(gem_oldclavier,gem_curclavier,sizeof(gem_oldclavier));
}

/*  */
/*  I_SetPalette */
/*  */
void I_SetPalette256_gem(byte *palette)
{
	int i,rgb[3];

	if (!vdi_index[16]) {
		for(i = 16; i < 255; i++)
			vdi_index[i] = i;
		vdi_index[255] = 1;
	}

	for(i = 0; i < 256; i++) {
		int r, g, b;

		r = gammatable[usegamma][*palette++];
		g = gammatable[usegamma][*palette++];
		b = gammatable[usegamma][*palette++];

		rgb[0] = (1000 * r) / 255;
		rgb[1] = (1000 * g) / 255;
		rgb[2] = (1000 * b) / 255;
		vs_color(vdi_handle, vdi_index[i], rgb);
	}
}

void I_InitGraphics_gem(void)
{
/*
	if (zoomscreen)
		I_ZoomInit(videowidth,videoheight);
*/
	/* PM: Bug in NOVA VDI on Atari TT : it reports 16 bits for 32K mode */
	/* PM: reported by Guillaume Tello */
	/* PM: so use '-video-15' to force 15 bits routines if 16 bits reported */

	if ((bpp==16) && M_CheckParm("-novavdi"))
		bpp=15;

	switch (bpp) {
		case 8:
			R_DrawColumn=R_DrawColumn8;
			R_DrawColumnLow=R_DrawColumnLow8;
			R_DrawFuzzColumn=R_DrawFuzzColumn8;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow8;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn8;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow8;
			R_DrawSpan=R_DrawSpan8;
			R_DrawSpanLow=R_DrawSpanLow8;
			AM_DrawFline=AM_drawFline8;
			wipe_doMelt=wipe_doMelt8;
			V_DrawPatch=V_DrawPatch8;
			V_DrawPatchFlipped=V_DrawPatchFlipped8;
			F_DrawPatchCol=F_DrawPatchCol8;
			I_Zoom=I_Zoom8;
			pixel_size=1;
			break;
		case 15:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			I_Zoom=I_Zoom16;
			fuzzmask=0x3DEF3DEF;
			break;
		case 16:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			I_Zoom=I_Zoom16;
			fuzzmask=0x7BEF7BEF;
			break;
		case 24:
			R_DrawColumn=R_DrawColumn24;
			R_DrawColumnLow=R_DrawColumnLow24;
			R_DrawFuzzColumn=R_DrawFuzzColumn24;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow24;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn24;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow24;
			R_DrawSpan=R_DrawSpan24;
			R_DrawSpanLow=R_DrawSpanLow24;
			AM_DrawFline=AM_drawFline24;
			wipe_doMelt=wipe_doMelt24;
			V_DrawPatch=V_DrawPatch24;
			V_DrawPatchFlipped=V_DrawPatchFlipped24;
			F_DrawPatchCol=F_DrawPatchCol24;
			pixel_size=3;
			I_Zoom=I_Zoom24;
			break;
		case 32:
			R_DrawColumn=R_DrawColumn32;
			R_DrawColumnLow=R_DrawColumnLow32;
			R_DrawFuzzColumn=R_DrawFuzzColumn32;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow32;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn32;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow32;
			R_DrawSpan=R_DrawSpan32;
			R_DrawSpanLow=R_DrawSpanLow32;
			AM_DrawFline=AM_drawFline32;
			wipe_doMelt=wipe_doMelt32;
			V_DrawPatch=V_DrawPatch32;
			V_DrawPatchFlipped=V_DrawPatchFlipped32;
			F_DrawPatchCol=F_DrawPatchCol32;
			I_Zoom=I_Zoom32;
			pixel_size=4;
			fuzzmask=0x007F7F7F;
			break;
	}
}

void I_SetMainScreen_gem(void)
{
	/* Directly draw to screen buffer */
	screens[0] = buffer;
}

static void wind_redraw(int winhandle, int *inside)
{
	int todo[4], flag;

	flag=1;
	while (!wind_update(BEG_UPDATE)) {
	}

	if (wind_get(wnd_handle,WF_FIRSTXYWH,&todo[0],&todo[1],&todo[2],&todo[3])) {

		if (rc_intersect((GRECT *)inside,(GRECT *)todo)) {
			todo[2]+=(todo[0]-1);
			todo[3]+=(todo[1]-1);
			vs_clip(vdi_handle,1,todo);
			copie_windowvdi(todo);
		}

		while (flag) {
			flag = wind_get(wnd_handle,WF_NEXTXYWH,&todo[0],&todo[1],&todo[2],&todo[3]);

			if ((todo[2]==0) || (todo[3]==0))
				flag=0;

			if (flag) {
				if (rc_intersect((GRECT *)inside,(GRECT *)todo)) {
					todo[2]+=(todo[0]-1);
					todo[3]+=(todo[1]-1);
					vs_clip(vdi_handle,1,todo);
					copie_windowvdi(todo);
				}
			}
		}
		vs_clip(vdi_handle,0,todo);
	}		

	wind_update(END_UPDATE);
}

static void copie_windowvdi(int *rect)
{
	MFDB	mfdb_src, mfdb_dest;
	int		pxy[8];

	wind_get(wnd_handle,WF_WORKXYWH,&pxy[0],&pxy[1],&pxy[2],&pxy[3]);

	mfdb_src.fd_addr=buffer;
	mfdb_src.fd_w=pxy[2];
	mfdb_src.fd_h=pxy[3];
	mfdb_src.fd_wdwidth=(mfdb_src.fd_w+15)>>4;
  	mfdb_src.fd_stand=0;   /* copie hard */
  	mfdb_src.fd_nplanes=bpp;
	mfdb_src.fd_r1=0;
  	mfdb_src.fd_r2=0;
  	mfdb_src.fd_r3=0;

	mfdb_dest.fd_addr=NULL; /* vers cran */

	pxy[0] = rect[0]-pxy[0];
	pxy[1] = rect[1]-pxy[1];
 	pxy[2] = pxy[0] + rect[2]-rect[0]+1;   
 	pxy[3] = pxy[1] + rect[3]-rect[1]+1;  

	pxy[4] = rect[0];
	pxy[5] = rect[1];
	pxy[6] = rect[2];  
	pxy[7] = rect[3];

	vro_cpyfm( vdi_handle, S_ONLY, pxy, &mfdb_src, &mfdb_dest);
}
